//
//  FolderListViewController.swift
//  StateDemoProject
//
//  Created by todd on 2021/7/30.
//

import UIKit
import TBaseFramework

class FolderListViewController: UIViewController, UITableViewDataSource,
                                StoreSubscriber,
                                EventHandler {
    
    @IBOutlet weak var tableView: UITableView!
    var previousFolder:AppState?

    override func viewDidLoad() {
        super.viewDidLoad()

        // Do any additional setup after loading the view.
        print( store.state.folderState.folders )
        tableView.register(UINib(nibName: "TableViewCell", bundle: nil), forCellReuseIdentifier: "cell")
        tableView.dataSource = self
        
        store.subscribe(self)
        EventCenter.shared.addEventHandler(eventHandler: self)
    }
    
    deinit {
        EventCenter.shared.removeEventHandler(eventHandler: self)
    }
    
    func newState(state: AppState) {
        if state.folderState.folders != previousFolder?.folderState.folders {
            tableView.reloadData()
        }
        previousFolder = state
    }
    
    func onEvent(event: Event) {
        if let e = event as? FolderDidAddEvent {
            tableView.insertRows(at: [IndexPath(item: e.position, section: 0)], with: .left)
        }
        else if let e = event as? FolderDidRemoveEvent {
            tableView.deleteRows(at: [IndexPath(item: e.position, section: 0)], with: .right)
        }
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "cell") as! TableViewCell
        
        if let folder = store.state.folderState.folders[indexPath.item] {
            cell.folderName.text = folder.name
            cell.folder = folder
        }
        
        return cell
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return store.state.folderState.folders.count
    }

}
