//
//  ViewController.swift
//  StateDemoProject
//
//  Created by todd on 2021/7/30.
//

import UIKit
import PureLayout
import TBaseFramework

class ViewController: UIViewController {

    override func viewDidLoad() {
        super.viewDidLoad()
        // Do any additional setup after loading the view.
        let root = UITestRootViewController(nibName: "UITestRootViewController", bundle: Bundle(for: UITestRootViewController.self))
        
        self.addChild(root)
        self.view.addSubview(root.view)
        root.view.configureForAutoLayout()
        root.view.autoPinEdgesToSuperviewEdges()
        root.didMove(toParent: self)
        
        initTestFolderListView()
    }

    func initTestFolderListView() {
        let dummyFolder = Folder(name: "Test folder")
        store.dispatch(AddFolderAction(folder: dummyFolder, position: 0))
        
        let testEntity = UITestEntry(title: "FolderListViewController") { viewContainer in
                       
            // create FolderListViewController and add it to test view
            let vc = FolderListViewController(nibName: "FolderListViewController", bundle: Bundle(for: FolderListViewController.self))
            
            viewContainer.addChild(vc)
            viewContainer.view.addSubview(vc.view)
            
            // for folder list view, we set it full size
            vc.view.configureForAutoLayout()
            vc.view.autoPinEdgesToSuperviewEdges()
            vc.didMove(toParent: viewContainer)
            
        }
        
        testEntity.actions.append(UITestAction(title: "Add Folder", actionHandler: {
            let newFolder = Folder(name: "New Folder")
            store.dispatch(AddFolderAction(folder: newFolder, position: 0))
            EventCenter.shared.send(event: FolderDidAddEvent(sender: self, newFolder: newFolder, position: 0))
        }))
        
        testEntity.actions.append(UITestAction(title: "Reload Folders", actionHandler: {
            let folders = [
                Folder(name: "Reload Folder 1"),
                Folder(name: "Reload Folder 2"),
                Folder(name: "Reload Folder 3"),
            ]
            store.dispatch(FoldersLoadedAction(folders: folders))
        }))
        
        testEntity.actions.append(UITestAction(title: "Delete Folder", actionHandler: {
            
            if let folder = store.state.folderState.folders[0] {
                store.dispatch(RemoveFolderAction(folder: folder))
                EventCenter.shared.send(event: FolderDidRemoveEvent(sender: self, folder: folder, position: 0))
            }

        }))
        
        // update State
        store.dispatch(UITestAddTestEntry(testEntry: testEntity))

    }

}


